@@ -695,7 +695,7 @@ action_authentications_editing (void)
   bool have_auth;
 
   AG_hide_all_frames ();
-  AG_insensitive_children ("anastasis_gtk_add_auth_button_box");
+  AG_insensitive_children ("anastasis_gtk_auth_button_grid");
   aps = json_object_get (AG_redux_state,
                          "authentication_providers");
   {
@@ -708,7 +708,8 @@ action_authentications_editing (void)
     {
       uint32_t ec = 0;
       uint32_t hc = 0;
-      json_t *methods;
+      bool disabled = false;
+      json_t *methods = NULL;
       struct GNUNET_JSON_Specification spec[] = {
         GNUNET_JSON_spec_mark_optional (
           GNUNET_JSON_spec_uint32 ("error_code",
@@ -716,8 +717,13 @@ action_authentications_editing (void)
         GNUNET_JSON_spec_mark_optional (
           GNUNET_JSON_spec_json ("methods",
                                  &methods)),
-        GNUNET_JSON_spec_uint32 ("http_status",
-                                 &hc),
+        GNUNET_JSON_spec_mark_optional (
+          GNUNET_JSON_spec_uint32 ("http_status",
+                                   &hc)),
+        GNUNET_JSON_spec_mark_optional (
+          GNUNET_JSON_spec_bool ("disabled",
+                                 &disabled)),
+
         GNUNET_JSON_spec_end ()
       };
 
@@ -729,6 +735,8 @@ action_authentications_editing (void)
         GNUNET_break (0);
         continue;
       }
+      if (disabled)
+        continue;
       switch (hc)
       {
       case MHD_HTTP_OK:
@@ -820,7 +828,7 @@ action_authentications_editing (void)
  * @param[out] method_cost cost to return
  * @return #GNUNET_OK on success
  */
-static int
+static enum GNUNET_GenericReturnValue
 lookup_recovery_cost (const char *provider,
                       const char *type,
                       struct TALER_Amount *method_cost)
@@ -1334,143 +1342,6 @@ action_secret_editing (void)
 }
 
 
-/**
- * Create the QR code image for our zone.
- *
- * @param scale factor for scaling up the size of the image to create
- * @param text text to encode
- * @return NULL on error
- */
-static GdkPixbuf *
-create_qrcode (unsigned int scale,
-               const char *text,
-               size_t text_size)
-{
-  QRinput *qri;
-  QRcode *qrc;
-  GdkPixbuf *pb;
-  guchar *pixels;
-  int n_channels;
-  const char *dir;
-  char *fn;
-  unsigned int size;
-
-  qri = QRinput_new2 (0, QR_ECLEVEL_M);
-  if (NULL == qri)
-  {
-    GNUNET_log_strerror (GNUNET_ERROR_TYPE_WARNING, "QRinput_new2");
-    return NULL;
-  }
-  /* first try encoding as uppercase-only alpha-numerical
-     QR code (much smaller encoding); if that fails, also
-     try using binary encoding (in case nick contains
-     special characters). */
-  if ((0 != QRinput_append (qri,
-                            QR_MODE_AN,
-                            text_size,
-                            (unsigned char *) text)) &&
-      (0 != QRinput_append (qri,
-                            QR_MODE_8,
-                            text_size,
-                            (unsigned char *) text)))
-  {
-    GNUNET_log_strerror (GNUNET_ERROR_TYPE_WARNING,
-                         "QRinput_append");
-    return NULL;
-  }
-  qrc = QRcode_encodeInput (qri);
-  if (NULL == qrc)
-  {
-    GNUNET_log_strerror (GNUNET_ERROR_TYPE_WARNING,
-                         "QRcode_encodeInput");
-    QRinput_free (qri);
-    return NULL;
-  }
-  /* We use a trick to create a pixbuf in a way that works for both Gtk2 and
-     Gtk3 by loading a dummy file from disk; all other methods are not portable
-     to both Gtk2 and Gtk3. */
-  dir = GNUNET_GTK_get_data_dir ();
-  GNUNET_asprintf (&fn,
-                   "%s%s",
-                   dir,
-                   "qr_dummy.png");
-  size = (qrc->width + 8) * scale;
-  size += 8 - (size % 8);
-  pb = gdk_pixbuf_new_from_file_at_size (fn,
-                                         size,
-                                         size,
-                                         NULL);
-  GNUNET_free (fn);
-  if (NULL == pb)
-  {
-    QRcode_free (qrc);
-    QRinput_free (qri);
-    return NULL;
-  }
-  pixels = gdk_pixbuf_get_pixels (pb);
-  n_channels = gdk_pixbuf_get_n_channels (pb);
-  for (unsigned int x = 4 * scale; x < size - 4 * scale; x++)
-    for (unsigned int y = 4 * scale; y < size - 4 * scale; y++)
-    {
-      unsigned int xx = x - 4 * scale;
-      unsigned int yy = y - 4 * scale;
-      unsigned int ss = size - 8 * scale;
-      unsigned int off =
-        (xx * qrc->width / ss) + (yy * qrc->width / ss) * qrc->width;
-      for (int c = 0; c < n_channels; c++)
-        pixels[(y * size + x) * n_channels + c] =
-          (0 == (qrc->data[off] & 1)) ? 0xFF : 0;
-    }
-  QRcode_free (qrc);
-  QRinput_free (qri);
-  return pb;
-}
-
-
-/**
- * Create the QR code image for our zone.
- *
- * @param text text to encode
- * @return NULL on error
- */
-static GdkPixbuf *
-setup_qrcode (const char *widget,
-              const char *text,
-              size_t text_size)
-{
-  GtkWidget *image;
-  GdkScreen *screen;
-  GtkSettings *settings;
-  gint dpi;
-  int scale;
-
-  image = GTK_WIDGET (GCG_get_main_window_object (widget));
-  if (NULL == image)
-  {
-    GNUNET_break (0);
-    return NULL;
-  }
-  /* adjust scale to screen resolution */
-  screen = gtk_widget_get_screen (GTK_WIDGET (image));
-  settings = gtk_settings_get_for_screen (screen);
-  g_object_get (G_OBJECT (settings),
-                "gtk-xft-dpi",
-                &dpi,
-                NULL);
-  if (-1 == dpi)
-    scale = 2;
-  else if (dpi >= 122800)
-    scale = 4;
-  else if (dpi >= 98304)
-    scale = 3;
-  else
-    scale = 2;
-  return create_qrcode (3 * scale,
-                        text,
-                        text_size);
-}
-
-
 static void
 action_truths_paying (void)
 {
@@ -1489,15 +1360,22 @@ action_truths_paying (void)
   {
     const char *payto = json_string_value (pt);
     GdkPixbuf *pb;
+    GtkWidget *w;
 
+    w = GTK_WIDGET (GCG_get_main_window_object ("unpaid_qr_treeview"));
+    if (NULL == w)
+    {
+      GNUNET_break (0);
+      continue;
+    }
     if (NULL == payto)
     {
       GNUNET_break (0);
       continue;
     }
-    pb = setup_qrcode ("unpaid_qr_treeview",
-                       payto,
-                       strlen (payto));
+    pb = AG_setup_qrcode (w,
+                          payto,
+                          strlen (payto));
     if (NULL == pb)
     {
       GNUNET_log (GNUNET_ERROR_TYPE_WARNING,
@@ -1569,7 +1447,14 @@ action_policies_paying (void)
       GNUNET_JSON_spec_end ()
     };
     GdkPixbuf *pb;
+    GtkWidget *w;
 
+    w = GTK_WIDGET (GCG_get_main_window_object ("unpaid_qr_treeview"));
+    if (NULL == w)
+    {
+      GNUNET_break (0);
+      continue;
+    }
     if (GNUNET_OK !=
         GNUNET_JSON_parse (ppr,
                            spec,
@@ -1578,9 +1463,9 @@ action_policies_paying (void)
       GNUNET_break (0);
       continue;
     }
-    pb = setup_qrcode ("unpaid_qr_treeview",
-                       payto,
-                       strlen (payto));
+    pb = AG_setup_qrcode (w,
+                          payto,
+                          strlen (payto));
     if (NULL == pb)
     {
       GNUNET_log (GNUNET_ERROR_TYPE_WARNING,
@@ -1978,9 +1863,13 @@ show_challenge_feedback (void)
     }
     if (NULL != taler_pay_uri)
     {
-      qr = setup_qrcode ("anastasis_gtk_challenge_status_treeview",
-                         taler_pay_uri,
-                         strlen (taler_pay_uri));
+      GtkWidget*w;
+
+      w = GTK_WIDGET (GCG_get_main_window_object (
+                        "anastasis_gtk_challenge_status_treeview"));
+      qr = AG_setup_qrcode (w,
+                            taler_pay_uri,
+                            strlen (taler_pay_uri));
     }
     if (TALER_EC_NONE != ec)
       emsg = TALER_ErrorCode_get_hint (ec);
@@ -2575,6 +2464,7 @@ action_challenge_paying (void)
       GNUNET_JSON_spec_end ()
     };
     GdkPixbuf *pb;
+    GtkWidget *w;
 
     if (GNUNET_OK !=
         GNUNET_JSON_parse (ppr,
@@ -2593,9 +2483,10 @@ action_challenge_paying (void)
                      "payment"))
       continue;
     found = true;
-    pb = setup_qrcode ("unpaid_qr_treeview",
-                       payto,
-                       strlen (payto));
+    w = GTK_WIDGET (GCG_get_main_window_object ("unpaid_qr_treeview"));
+    pb = AG_setup_qrcode (w,
+                          payto,
+                          strlen (payto));
     if (NULL == pb)
       GNUNET_log (GNUNET_ERROR_TYPE_WARNING,
                   _ ("Failed to initialize QR-code pixbuf for `%s'\n"),
@@ -2650,6 +2541,15 @@ action_challenge_paying (void)
  * Render challenge feedback for challenge @a uuid_str in a dialog of
  * @a builder in the label under @a target_widget.
  *
+ * Useful in case the operation previously failed at the
+ * server and we have some useful information to return
+ * to the user.
+ *
+ * FIXME: only do it AFTER the first attempt of the
+ * user to enter a code, and/or change what the
+ * server returns so we do NOT render a confusing
+ * error message on first use!
+ *
  * @param builder a builder to get widgets from
  * @param target_widget the widget to upate
  * @param uuid_str the UUID to render feedback for
@@ -2668,6 +2568,7 @@ render_feedback (GtkBuilder *builder,
   const char *taler_pay_uri = NULL;
   uint32_t ec = 0;
   uint32_t http_status = 0;
+  bool hide = false;
   struct GNUNET_JSON_Specification spec[] = {
     GNUNET_JSON_spec_string ("state",
                              &state),
@@ -2763,7 +2664,20 @@ render_feedback (GtkBuilder *builder,
     else
     {
       const char *ihint;
+      const char *type = "Error";
 
+      switch (code)
+      {
+      case TALER_EC_ANASTASIS_TRUTH_CHALLENGE_RESPONSE_REQUIRED:
+        hide = true;
+        type = "Hint";
+        break;
+      case TALER_EC_ANASTASIS_TRUTH_CHALLENGE_ACTIVE:
+        type = "Hint";
+        break;
+      default:
+        break;
+      }
       ihint = TALER_ErrorCode_get_hint (code);
       if ( (NULL != hint) &&
            ( (NULL == ihint) ||
@@ -2775,11 +2689,13 @@ render_feedback (GtkBuilder *builder,
       {
         if (NULL == ihint)
           GNUNET_asprintf (&msg,
-                           "Error #%u",
+                           "%s #%u",
+                           type,
                            (unsigned int) code);
         else
           GNUNET_asprintf (&msg,
-                           "Error #%u: %s",
+                           "%s #%u: %s",
+                           type,
                            (unsigned int) code,
                            ihint);
       }
@@ -2787,12 +2703,14 @@ render_feedback (GtkBuilder *builder,
       {
         if (NULL == ihint)
           GNUNET_asprintf (&msg,
-                           "Error #%u (%s)",
+                           "%s #%u (%s)",
+                           type,
                            (unsigned int) code,
                            detail);
         else
           GNUNET_asprintf (&msg,
-                           "Error #%u: %s (%s)",
+                           "%s #%u: %s (%s)",
+                           type,
                            (unsigned int) code,
                            ihint,
                            detail);
@@ -2809,7 +2727,10 @@ render_feedback (GtkBuilder *builder,
   gtk_label_set_text (elabel,
                       msg);
   GNUNET_free (msg);
-  gtk_widget_show (GTK_WIDGET (elabel));
+  if (hide)
+    gtk_widget_hide (GTK_WIDGET (elabel));
+  else
+    gtk_widget_show (GTK_WIDGET (elabel));
   GNUNET_JSON_parse_free (spec);
 }
 
@@ -2852,7 +2773,6 @@ diag_question (const json_t *details)
 
     uuid_str = json_string_value (json_object_get (details,
                                                    "uuid"));
-    /* Why do we do this? */
     render_feedback (builder,
                      "anastasis_gtk_c_question_error_label",
                      uuid_str);
@@ -2911,7 +2831,6 @@ diag_code (const json_t *details)
                                                "anastasis_gtk_c_challenge_label"));
     gtk_label_set_text (label,
                         uuid_str);
-    /* Why do we do this? */
     render_feedback (builder,
                      "anastasis_gtk_c_code_error_label",
                      uuid_str);
@@ -2926,6 +2845,66 @@ diag_code (const json_t *details)
 }
 
 
+/**
+ * Create a dialog for the user to enter an TOTP code.
+ *
+ * @param details details about the dialog to render
+ * @return dialog object
+ */
+static GtkDialog *
+diag_totp (const json_t *details)
+{
+  GtkBuilder *builder;
+  const char *instructions;
+  const char *uuid_str;
+  struct GNUNET_JSON_Specification spec[] = {
+    GNUNET_JSON_spec_string ("instructions",
+                             &instructions),
+    GNUNET_JSON_spec_string ("uuid",
+                             &uuid_str),
+    GNUNET_JSON_spec_end ()
+  };
+
+  if (GNUNET_OK !=
+      GNUNET_JSON_parse (details,
+                         spec,
+                         NULL, NULL))
+  {
+    GNUNET_break (0);
+    json_dumpf (details,
+                stderr,
+                JSON_INDENT (2));
+    return NULL;
+  }
+
+  builder = GNUNET_GTK_get_new_builder ("anastasis_gtk_challenge_totp.glade",
+                                        NULL);
+  if (NULL == builder)
+  {
+    GNUNET_break (0);
+    return NULL;
+  }
+  {
+    GtkLabel *label;
+
+    label = GTK_LABEL (gtk_builder_get_object (builder,
+                                               "challenge_instructions_label"));
+    gtk_label_set_text (label,
+                        instructions);
+  }
+  render_feedback (builder,
+                   "anastasis_gtk_c_totp_error_label",
+                   uuid_str);
+  {
+    GtkDialog *ad;
+
+    ad = GTK_DIALOG (gtk_builder_get_object (builder,
+                                             "anastasis_gtk_c_totp_dialog"));
+    return ad;
+  }
+}
+
+
 /**
  * Create a dialog for the user to make an IBAN transfer.
  *
@@ -3058,6 +3037,8 @@ action_challenge_solving (void)
       .ctor = &diag_code },
     { .type = gettext_noop ("iban"),
       .ctor = &diag_iban },
+    { .type = gettext_noop ("totp"),
+      .ctor = &diag_totp },
     { .type = NULL,
       .ctor = NULL }
   };
@@ -3104,11 +3085,11 @@ action_challenge_solving (void)
   /* show dialog */
   {
     GtkWidget *toplevel;
-    GtkBox *box;
+    GtkWidget *box;
 
-    box = GTK_BOX (GCG_get_main_window_object (
-                     "anastasis_gtk_open_challenge_box"));
-    toplevel = gtk_widget_get_toplevel (GTK_WIDGET (box));
+    box = GTK_WIDGET (GCG_get_main_window_object (
+                        "anastasis_gtk_open_challenge_box"));
+    toplevel = gtk_widget_get_toplevel (box);
     gtk_window_set_transient_for (GTK_WINDOW (diag),
                                   GTK_WINDOW (toplevel));
     gtk_window_present (GTK_WINDOW (diag));
@@ -3179,22 +3160,22 @@ action_recovery_finished (void)
     AG_show ("anastasis_gtk_secret_copy_button");
   }
   pb = NULL;
+  img = GTK_IMAGE (GCG_get_main_window_object (
+                     "anastasis_gtk_secret_qr_image"));
   if (NULL != text)
   {
-    pb = setup_qrcode ("anastasis_gtk_secret_qr_image",
-                       text,
-                       strlen (text));
+    pb = AG_setup_qrcode (GTK_WIDGET (img),
+                          text,
+                          strlen (text));
   }
   else
   {
-    pb = setup_qrcode ("anastasis_gtk_secret_qr_image",
-                       data,
-                       data_size);
+    pb = AG_setup_qrcode (GTK_WIDGET (img),
+                          data,
+                          data_size);
   }
   if (NULL != pb)
   {
-    img = GTK_IMAGE (GCG_get_main_window_object (
-                       "anastasis_gtk_secret_qr_image"));
     gtk_image_set_from_pixbuf (img,
                                pb);
     g_object_unref (pb);
