@@ -28,6 +28,8 @@
 #include <gnunet/gnunet_util_lib.h>
 #include "anastasis-gtk_helper.h"
 #include <jansson.h>
+#include <qrencode.h>
+#include <gdk-pixbuf/gdk-pixbuf.h>
 
 
 /**
@@ -263,3 +265,132 @@ AG_error (const char *format,
   AG_have_error = true;
   gtk_widget_show (GTK_WIDGET (l));
 }
+
+
+/**
+ * Create a the QR code image from a given @a text.
+ *
+ * @param scale factor for scaling up the size of the image to create
+ * @param text text to encode
+ * @return NULL on error
+ */
+static GdkPixbuf *
+create_qrcode (unsigned int scale,
+               const char *text,
+               size_t text_size)
+{
+  QRinput *qri;
+  QRcode *qrc;
+  GdkPixbuf *pb;
+  guchar *pixels;
+  int n_channels;
+  const char *dir;
+  char *fn;
+  unsigned int size;
+
+  qri = QRinput_new2 (0, QR_ECLEVEL_M);
+  if (NULL == qri)
+  {
+    GNUNET_log_strerror (GNUNET_ERROR_TYPE_WARNING, "QRinput_new2");
+    return NULL;
+  }
+  /* first try encoding as uppercase-only alpha-numerical
+     QR code (much smaller encoding); if that fails, also
+     try using binary encoding (in case nick contains
+     special characters). */
+  if ((0 != QRinput_append (qri,
+                            QR_MODE_AN,
+                            text_size,
+                            (unsigned char *) text)) &&
+      (0 != QRinput_append (qri,
+                            QR_MODE_8,
+                            text_size,
+                            (unsigned char *) text)))
+  {
+    GNUNET_log_strerror (GNUNET_ERROR_TYPE_WARNING,
+                         "QRinput_append");
+    return NULL;
+  }
+  qrc = QRcode_encodeInput (qri);
+  if (NULL == qrc)
+  {
+    GNUNET_log_strerror (GNUNET_ERROR_TYPE_WARNING,
+                         "QRcode_encodeInput");
+    QRinput_free (qri);
+    return NULL;
+  }
+  /* We use a trick to create a pixbuf in a way that works for both Gtk2 and
+     Gtk3 by loading a dummy file from disk; all other methods are not portable
+     to both Gtk2 and Gtk3. */
+  dir = GNUNET_GTK_get_data_dir ();
+  GNUNET_asprintf (&fn,
+                   "%s%s",
+                   dir,
+                   "qr_dummy.png");
+  size = (qrc->width + 8) * scale;
+  size += 8 - (size % 8);
+  pb = gdk_pixbuf_new_from_file_at_size (fn,
+                                         size,
+                                         size,
+                                         NULL);
+  GNUNET_free (fn);
+  if (NULL == pb)
+  {
+    QRcode_free (qrc);
+    QRinput_free (qri);
+    return NULL;
+  }
+  pixels = gdk_pixbuf_get_pixels (pb);
+  n_channels = gdk_pixbuf_get_n_channels (pb);
+  for (unsigned int x = 4 * scale; x < size - 4 * scale; x++)
+    for (unsigned int y = 4 * scale; y < size - 4 * scale; y++)
+    {
+      unsigned int xx = x - 4 * scale;
+      unsigned int yy = y - 4 * scale;
+      unsigned int ss = size - 8 * scale;
+      unsigned int off =
+        (xx * qrc->width / ss) + (yy * qrc->width / ss) * qrc->width;
+      for (int c = 0; c < n_channels; c++)
+        pixels[(y * size + x) * n_channels + c] =
+          (0 == (qrc->data[off] & 1)) ? 0xFF : 0;
+    }
+  QRcode_free (qrc);
+  QRinput_free (qri);
+  return pb;
+}
+
+
+GdkPixbuf *
+AG_setup_qrcode (GtkWidget *w,
+                 const char *text,
+                 size_t text_size)
+{
+  GdkScreen *screen;
+  GtkSettings *settings;
+  gint dpi;
+  int scale;
+
+  if (NULL == w)
+  {
+    GNUNET_break (0);
+    return NULL;
+  }
+  /* adjust scale to screen resolution */
+  screen = gtk_widget_get_screen (w);
+  settings = gtk_settings_get_for_screen (screen);
+  g_object_get (G_OBJECT (settings),
+                "gtk-xft-dpi",
+                &dpi,
+                NULL);
+  if (-1 == dpi)
+    scale = 2;
+  else if (dpi >= 122800)
+    scale = 4;
+  else if (dpi >= 98304)
+    scale = 3;
+  else
+    scale = 2;
+  return create_qrcode (3 * scale,
+                        text,
+                        text_size);
+}
